package schema;

import java.io.*;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

/**
 * DTDParser parses a DTD file and extracts the set of semantic tags
 * defined in it.  
 * 
 * DTD file desribes the structure of the semantic tags allowed.
 * Each line in the DTD file contains one of the following:
 * 
 * <ol>
 *   <li> an element declaration in the format
 *		<!ELEMENT el-name (el-category)> OR
 *	 	<!ELEMENT el-name (element-content)>
 *	Such a declaration may continue on one or more consecutive 
 *	lines in the file.
 *   </li>
 * 
 *   <li> an attribute declaration, which is currently discarded
 * 		<!ATTLIST el-name attribute-name attribute-type default-value>
 *   </li>
 * 
 *   <li> an element comment; it contains the name of the element to be
 *		annotated and its meaning (separated with a space); at the end
 *		"." has to be present, otherwise the comment won't be processed.
 *		<!-- el-name comment. -->
 *   </li>
 * 
 *   <li> other comments
 *		<!-- some other comment with no strict format -->
 *   </li>
 * 
 * @author Stani Vlasseva
 */

public class DTDParser
{
	// name of the corresponding DTD file
	private String dtdFileName;
	
	// name of the file specifying the top-level tags
	private String topLevelTagsFileName;
	
	// semantic tag set (a hashtable of SemanticTag objects with key the tag names)
	private Hashtable tagSet;
	
	/**
	 * Creates a new DTDParser object for a given DTD file
	 * 
	 * @param dtdFileName the name of the DTD file
	 */
	public DTDParser(String dtdFileName, String topLevelTagsFileName) {
		this.dtdFileName = dtdFileName;
		this.topLevelTagsFileName = topLevelTagsFileName;
								
		this.tagSet = new Hashtable();
		// populate the tagSet
		try {
			this.tagSet = this.extractTagSet();
		}
		catch (IOException e) {
			System.err.println(e);
		}		
	}
	
	/**
	 * Returns the name of the dtd file associated with that DTDParser object.
	 * 
	 * @return dtdFileName name of the DTD file
	 */
	public String getDTDFileName() {
		return this.dtdFileName;
	}
	
	/**
	 * Returns the name of the text file containig the top-level tags.
	 * 
	 * @return name of the file containing the top-level tags (i.e., 
	 * the "root" children).
	 */
	public String getTopLevelTagsFileName() {
		return this.topLevelTagsFileName;
	}
	
	/**
	 * Returns the semantic tag set extracted from the DTD file.
	 * 
	 * @return semantic tag set
	 */
	public Hashtable getTagSet() {
		return this.tagSet;
	}
	
	/**
	 * Returns a hashtable, which contains all annotated semantic tags.
	 * Tag names are used as keys in the hashtable and their comments
	 * are the corresponding values.
	 * 
	 * @return hashtable containing all annotated tags
	 */
	public Hashtable getComments() {
		SemanticTag tag;
		Hashtable tagSet = this.getTagSet();		
		Hashtable comments = new  Hashtable();
		
		Enumeration keys = this.getTagSet().keys();
		
		while ( keys.hasMoreElements() ) {
			String key = (String) keys.nextElement();
	
			// skip the artificial "root" element
			// and put all the rest tag names with
			// their corresponding meanings in the hashtable
			if ( !key.equals("root") ) {
				tag = (SemanticTag) tagSet.get(key);
			
				comments.put(key, tag.getComment());
			}
		}	
		return comments;
	}		
	
	/**
	 * Extracts the tag set from the DTD file. 
	 */		
	private Hashtable extractTagSet() throws IOException {						
		FileInputStream input = new FileInputStream(this.getDTDFileName());		
		InputStreamReader reader = new InputStreamReader(input);
		BufferedReader buffer = new BufferedReader(reader);
		
		Hashtable set = new Hashtable();
				
		String line, element;	
		String description;
		SemanticTag tag, childTag;
		
		/*
		In the DTD file the following types of representations may apper:
		
		(1) an element declaration in the format:
			<!ELEMENT el-name (el-category)> OR
	 		<!ELEMENT el-name (element-content)>
			Such a declaration may continue on one or more consecutive 
			lines in the file.
		
		(2) an attribute declaration which is currently discarded
			<!ATTLIST el-name attribute-name attribute-type default-value>
		
		(3) an element comment; it contains the name of the element to be
			annotated and its meaning (separated with a space); at the end
			"." has to be present, otherwise the comment won't be processed.
			<!-- el-name comment. -->
		
		(4) other comments
			<!-- some other comments with no strict format -->
		*/
		
		// for each line in the input file
		while ( (line = buffer.readLine()) != null ) {
			
			// case (1) from the above
			if ( line.startsWith("<!ELEMENT") ) {
				
				// if the line contains a complete element declaration
				if ( line.endsWith(">") ) {
					element = line.trim();
					
					// process the element
					processElement(element, set);
					
					// extract the next line from the input file
					continue;
				}
				
				else {
					// an element continuing on the consecutive line(s)
					element = line.trim();
					
					// concatinate the whole declaration
					while ( !(line = buffer.readLine()).endsWith(">") ) {
						element = element.concat(" ").concat(line.trim());
					}
					element = element.concat(" ").concat(line.trim());
										
					//process the element
					processElement(element, set);
					
					// extract the next line from the input file
					continue;
				}								
			}
			
			else {
				// if the line contains some kind of a comment case (3) or (4)
				if ( line.startsWith("<!--") ) {
					String comment;
					
					//if the comment ends on the same line
					if ( line.endsWith("-->") ) {
						comment = line.trim();
						
						// process the comment
						processComment(comment, set);						
						
						// extract the next line from the input file
						continue;
					}
					else {
						// a comment continuing on the next line(s)
						comment = line.trim();
						
						// concatinate the whole comment in a single line
						while ( !(line = buffer.readLine()).endsWith(">") ) {
							comment = comment.concat(" ").concat(line.trim());
						}
						comment = comment.concat(" ").concat(line.trim());
					
					//process the comment
					processComment(comment, set);
					
					// extract the next line from the input file
					continue;
					}															
				}
				
				else {
					// it should be case (2) from above, i.e., attribute definition
					// skip it for now
					continue;
				}
			}
			
		}
		input.close();
				
		/* Create an artificial root node with parent "" and children 
		   all elements specified in the text file with top-level elements
		*/
		
		// read the root children from a given txt file
		input = new FileInputStream(this.getTopLevelTagsFileName());		
		reader = new InputStreamReader(input);
		buffer = new BufferedReader(reader);
		
		Vector rootChildren = new Vector();
		Enumeration keys = set.keys();		
				
		while ( (line = buffer.readLine()) != null ) {
			rootChildren.addElement(line.trim());
		}
		
		while ( keys.hasMoreElements() ) {
			String key = (String) keys.nextElement();
			
			if ( rootChildren.contains(key) ) {
				tag = (SemanticTag) set.get(key);
				tag.addParent("root");
				
				set.remove(key);
				set.put(key, tag);
			}
		}
		
		// create the root element and add it to the tag set
		SemanticTag root = new SemanticTag("root", "", rootChildren);
		root.addComment("Root of the tag set hierarchy.");
		set.put("root", root);
		
		// close the file with top-level elements
		input.close();	
		
		return set;
	}
	
	/**
	 * Processes an element writen in DTD format and puts it in
	 * a hashtable with a key the name of the element and its
	 * representation (as a SemanticTag object) as a corresponding value.
	 * 
	 * @param element DTD element declaration to be processed
	 * @param table hashtable where the result after the parsing is stored
	 */
	private void processElement(String element, Hashtable table) {		
			Vector children = new Vector();	// list of children names as strings
											// the accumulation of all children for the 
											// current element
			
			int nextSpace, lastParenthesis;
			String tagName;
			String word;
			
			// find the index of the first space (the one after "<!ELEMENT")
			nextSpace = element.indexOf(" ");
			if ( nextSpace == -1 ) {
				System.err.println("Incorrect element format...");
				System.exit(1);
			}
			
			// find the index of the first ")" looking backwards
			lastParenthesis = element.lastIndexOf(")");		
			
			// extract the string starting after the " " to the ")"; 
			// add " " at the end to be able to treat the last child 
			// the way all previous ones are treated
			element = element.substring(nextSpace + 1, lastParenthesis).concat(" ");				
			
			// find the next space and extract the element (sem tag) name
			nextSpace = element.indexOf(" ");
			tagName = element.substring(0, nextSpace);
			//System.out.println("name: " + tagName);
			
			// remove the element name from the string;
			// + 2 to skip the " (" after the name
			element = element.substring(nextSpace + 2);	
			
			/* At that point only the children or the category (type)
			 * the tag are left in the line.
			 * Extract each child one by one and do one of the following:
			 * (1) If there is no tag with that name already in the tag set
			 *	   create a SemanticTag object for that child with parent
			 *	   the current tagName (i.e., the element name) and add it
			 *	   to the tag set;
			 * (2) If a tag with the same name is found in the set, then 
			 *     add only the current tagName as his parent in the list of
			 *     parents.
			 *  In both cases the child is added to the list of children for
			 *  the current tagName.
			 */						
			while ( element.length() > 0 ) {
				nextSpace = element.indexOf(" ");
				if ( nextSpace == -1 )
					break;
				word = element.substring(0, nextSpace);
				
				// check if the read word is a legitimate child
				// if it starts with "#" its a type, not a real child
				if ( !word.equals("|") && (!word.startsWith("#")) ) {
					
					// if the word ends with a "?", "+", "*", or "," 
					// remove them before storing it
					if ( word.endsWith("?") || word.endsWith("+") || 
						 word.endsWith("*") || word.endsWith(",") ) {
						word = word.substring(0, word.length()-1 );
					}
					
					children.addElement(word);
					
					SemanticTag present = (SemanticTag) table.get(word);
					if ( present == null ) {
						// the element is not in the tag set yet
						SemanticTag childTag = new SemanticTag(word, tagName);
						table.put(word, childTag);
					}
					else {
						// the element is already present								
						present.addParent(tagName);	

						table.remove(word);
						table.put(word, present);
					}
					
					// when the end of the line is reached 
					// go to read the next line from the file
					if ( element.length() > nextSpace + 1 )		// ??? >=
						element = element.substring(nextSpace + 1);
					else {
						break;
					}
				}
				else {
					element = element.substring(nextSpace + 1);
				}
			}
							
			/* After all children have been read and the info for their
			 * paren is added in the tag set create a SemanticTag object for 
			 * the element itself (i.e., the element which declaration was written 
			 * on that line).
			 * Again, there are two possible cases:
			 * (1) If a tag with the same name is already in the tag set, then add
			 *     only its children; (This means that the current element has appeared
			 *     earlier as a child of another element.)
			 * (2) If the tag is not in the tag set already (this is its first appearance
			 *     in the file) add it to the tag set with all its children, and without 
			 *     a parent.
			*/
			SemanticTag tag = new SemanticTag(tagName, children);
			SemanticTag present = (SemanticTag) table.get(tagName);
			
			if ( present == null ) {
				// the element is not in the set yet
				table.put(tagName, tag);
			}
			
			else {
				// this is the definition of an element wchich is already present in 
				// the tag set, because it has appeared as a child earlier
				present.addChildren(children);
				
				table.remove(tagName);
				table.put(tagName, present);
			}					
	}				
	
	/**
	 * Processes a DTD comment.
	 * 
	 * @param comment DTD comment to be processed
	 * @param table hashtable there the parsed information is stored
	 */
	private void processComment(String comment, Hashtable table) {
		
		// skip the "<!--" at the beginning and 
		// "-->" at the end; remove all the space at the beginning and the end
		comment = comment.substring(4, (comment.length() - 3)).trim();	
		
		if ( !comment.endsWith(".") ) {
			// case (4), i.e., comment not relevant to the tag meaning
			return;
		}
		else {
			// case (3)
			// The comment contains an element and its annotation in the format
			// '<element-name> <comments>.'
			int nextSpace;
			String element;
			String description;
					
			nextSpace = comment.indexOf(" ");
			if ( nextSpace == -1 ) {
				System.out.println("Incorrect comment format...");
				System.exit(1);
			}
			
			element = comment.substring(0, nextSpace);
			description = comment.substring(nextSpace + 1);
			
			SemanticTag tag = (SemanticTag) table.get(element);
			if ( tag != null ) {
				// the element is already in the set
				// just add its meaning
				tag.addComment(description);

				table.remove(element);
				table.put(element, tag);
			}
			else {
				// the meaning of the tag apperas before its definition
				// create a new entry into the tag set
				SemanticTag newTag = new SemanticTag(element);
				newTag.addComment(description);
				table.put(element, newTag);
			}
		}
	}
	
	/**
	 *  Traverses the semantic tags from the tree with root a tag named tagName.
	 *  The result is printed to the standart output; the different levels 
	 *  in the hieararchy are indented.
	 * 
	 * @param tagName name of the tag at which traversal will start
	 * @param depth the level in the hierarchical structure of tags to be
	 *				assosiated with that tag
	 * */
	private void traverse(String tagName, int depth) {
		SemanticTag tag;
		int level = depth;
		int numberOfChildren;
		
		// check if that tag is in the tag set
		tag = (SemanticTag) this.getTagSet().get(tagName);
		if ( tag == null ) {
			System.err.println("There's no tag <" + tagName + "> in the tag set!");
			System.exit(1);
		}		
		else {
			// the element is found
			numberOfChildren = tag.getChildrenTags().size();
			
			int counter = level;
			// put a new line before each element from the top level
			if ( counter == 1)
				System.out.println();
			while ( counter >=  0) {		
				System.out.print("    ");
				counter--;
			}
			if ( !tagName.equals("root") )		// to skip the "root"
				System.out.println(tagName);		
			
			if ( numberOfChildren > 0 ) {			
				// there are some children
				for ( int k=0; k < numberOfChildren; k++ ) {
					String childName = (String)tag.getChildrenTags().elementAt(k);						
					traverse(childName, level + 1);
				}
			}
		}
	}			
}
