package schema;

import java.io.*;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;

/**
 * Generates a tree of HTML documents, which represents the structure of
 * the semantic tag set.
 * 
 * @author Stani Vlasseva
 */
public class HTMLTreeGenerator
{
	// name of the dtd file from which the tree is generated
	private String dtdFileName;
	
	// name of the file specifying the top-level tags
	private String topLevelTagsFileName;
	
	private String version;
		// version number of the html tree
	
	/**
	 * Creates an HTMLTreeGenerator object for a given DTD file.
	 * 
	 * @param dtdFileName name of the DTD file from which the HTML tree is extracted
	 */
	public HTMLTreeGenerator(String dtdFileName, String topLevelTagsFileName, String version) {
		this.dtdFileName = dtdFileName;
		this.topLevelTagsFileName = topLevelTagsFileName;
		this.version = version;
		
		try {
			this.createHTMLTree();
		}
		catch (IOException e) {
			System.err.println(e);
		}
	}
	
	/**
	 * Returns the name of the dtd file associated with the HTMLTreeGenerator object.
	 * 
	 * @return dtdFileName name of the DTD file
	 */
	public String getDTDFileName() {
		return this.dtdFileName;
	}
	
	/**
	 * Returns the name of the text file containig the top-level tags.
	 * 
	 * @return name of the file containing the top-level tags (i.e., 
	 * the "root" children).
	 */
	public String getTopLevelTagsFileName() {
		return this.topLevelTagsFileName;
	}
	
	/**
	 * Returns the version number of the html tree.
	 * @return version number
	 */
	public String getVersion() {
		return this.version;
	}
	
	/**
	 * createHTMLTree() is the method which acctually creates the tree
	 * associated with an object from the HTMLTreeGenerator class.
	 */
	private void createHTMLTree() throws IOException {
		DTDParser parser = new DTDParser(this.getDTDFileName(), this.getTopLevelTagsFileName());
		Hashtable tagSet = parser.getTagSet();
		Enumeration keys = tagSet.keys();
		SemanticTag tag;
		String key;
		String fileName;	// name of the html file to be created			
		
		while ( keys.hasMoreElements() ) {
			key = (String) keys.nextElement();
			
			if ( key.equals("root") ) {
				// for the root element create an html file called "index.html"
				fileName = "index.html";
				tag = (SemanticTag) tagSet.get(key);
			}
			else {
				// for the rest elemetns if they have children create
				// an html file called <name of the tag>.html;
				// if the elements are leaf tags: don't create a separate html file
				
				tag = (SemanticTag) tagSet.get(key);
				
				// if the tag has nested tags different from <value> and <ignore>
				if ( tag.getChildrenTags().size() > 2 ) {
					fileName = key.concat(".html");
				}
				else
					// the element has no children, or its children are onle <value>
					// and <ignore>; do not create a separate html file for that element.
					continue;
			}
			
			// open the file to write
			FileOutputStream output = new FileOutputStream(fileName);
			PrintWriter writer = new PrintWriter(output);
			
			writer.println("<!-- This file has been created automatically from the \"" +
						   this.getDTDFileName() + "\" file. -->\n");
			writer.println("<html>");
			
			writer.println("</html>");
			writer.println("<head>");
			writer.println("<title>UW Mangrove Project: Schema Version " + this.getVersion() + "</title>");
			writer.println("</head>\n");

			writer.println("<body>");
			writer.println("<center>");
			writer.println("<h1><a href=\"../index.html\">UW Mangrove Project</a></h1>");
			writer.println("<h2>Schema Version " + this.getVersion() + "</h2>");
			writer.println("</center>");
			
			writer.println();
			
			// add the name of the parent tag
			if ( !tag.getName().equals("root")) {				
				writer.println("<a href=\"./index.html\"><b>Top-level Tags</b></a><br><br>");
				writer.println("<a href=\"../annotation.html\"><b>Annotation Example</b></a>");
				writer.println("<hr widht=75%>\n");
				writer.println("<h2>Tags permitted inside a <b>&lt;uw:" + tag.getName() + "&gt;</b> tag:</b></h2>");
			}
			else {
				// for the top-level tags
				writer.println("<a href=\"../annotation.html\"><b>Annotation Example</b></a>");
				writer.println("<hr widht=75%>\n");
				writer.println("<h2>Valid top-level tags:</h2>");
			}
			//writer.println("<hr widht=75%>\n");
			
			writer.println("<p>");
			writer.println("<dl>");
			
			// for each child
			Vector children = tag.getChildrenTags();
			String childKey;
			SemanticTag childTag;
			
			for ( int i=0; i<children.size(); i++ ) {
				
				writer.println("  <p>");
				
				childKey = (String) children.elementAt(i);
				childTag = (SemanticTag) tagSet.get(childKey);
				
				// if the child has no children, ot if it has only 
				// <value> and <ignore> as its children, treat it as 
				// a leaf node (i.e., do not include them in the html 
				// file for that element).
				if ( childTag.getChildrenTags().size() <= 2 ) {
					// leaf tag
					if ( !childKey.equals("value") && !childKey.equals("ignore") ) {
						writer.println("  <dt><b>&lt;uw:" + childTag.getName() + "&gt;</b></dt>");	
						// write the tag comment
						writer.println("  <dd>" + childTag.getComment() + "</dd>");
					}
				}
				
				else {
					// comlex tag, i.e., it has children
					String childFileName = childTag.getName().concat(".html");
					writer.println("  <dt><b>&lt;<a href=\"./" + childFileName + "\">uw:" + childTag.getName() + "</a>&gt;</b></dt>");					
					// write the tag comment
					writer.println("  <dd>" + childTag.getComment() + "</dd>");
				}
				
				writer.println("  </p>\n");				
			}
			
			writer.println("</dl>");
			writer.println("</p>\n");							
			
			writer.println("</body>");
			writer.println("</html>");
			
			writer.flush();
			writer.close();
		}
	}
}
