package schema;

import java.util.Vector;

/**
 * SemanticTag class represents an element from a DTD file (the element itself
 * describes a semantic tag). It has a name, a list of parents, a list of children, 
 * and a comment about it's meaning.
 * 
 * @author Stani Vlasseva
 */
public class SemanticTag
{
	private String name;		// tag name
	private Vector parents;		// a list of all possible parents in the hierarchy;
								// If the semantic tag is at the top level in the
								// hierarchy its parent is an empty string at first
								// position.
	private Vector childrenTags;// a list of all allowable children for complex tags
								// i.e., for tags that have nested tags; for simple type 
								// tags the list is empty;
	private String comment;		// comment describing the meaning of the tag
	
	/** Default constructor without any arguments.
	 * Creates an "empty" semanticTag.
	 */
	public SemanticTag(){
		this.name = "";
		this.parents = new Vector();
		this.childrenTags = new Vector();
		this.comment = "";
	}
	
	/**
	 * Creates a SemanticTag instance with a given name.
	 * 
	 * @param name name of the semantic tag.
	 */
	public SemanticTag(String name) {
		this.name = name;
		this.parents = new Vector();
		this.childrenTags = new Vector();
		this.comment = "";
	}
	
	/**
	 * Creates a SemanticTag object with a given name and a parent.
	 * 
	 * @param name name of the semantic tag
	 * @param parent name of its parent tag
	 */
	public SemanticTag(String name, String parent){
		this.name = name;
		this.parents = new Vector();
		this.parents.addElement(parent);
		this.childrenTags = new Vector();
		this.comment = "";
	}
	
	/**
	 * Creates a SemanticTag object with a given name, parent and a list of children.
	 * 
	 * @param name name of the semantig tag
	 * @param parent name of the parent tag
	 * @param children a list of all children tags allowed for that tag
	 */
	public SemanticTag(String name, String parent, Vector childrenTags) {
		this.name = name;
		this.parents = new Vector();
		this.parents.addElement(parent);
		this.childrenTags = new Vector();
		this.childrenTags = childrenTags;
		this.comment = "";
	}
	
	/**
	 * Creates a SemanticTag instance with a given name and a list of children
	 * 
	 * @param name the name of the semantic tag
	 * @param childrenTags a list of allowed children tags names
	 */
	public SemanticTag(String name, Vector childrenTags) {
		this.name = name;
		this.childrenTags = new Vector();
		this.childrenTags = childrenTags;
		this.parents = new Vector();
		this.comment = "";
	}
		
	/**
	 * Creates an instance of the SemanticTag class with a given name,
	 * a list of parents and a list of children.
	 * 
	 * @param name the name of the semantic tag
	 * @param parents a list of all possible parents
	 * @param children a list of all possible children
	 */
	public SemanticTag(String name, Vector parents, Vector children) {
		this.name = name;
		this.parents = new Vector();
		this.parents = parents;
		this.childrenTags = new Vector();
		this.childrenTags = children;
		this.comment = "";
	}
	
	// SemanticTag methods
	
	/**
	 * Returns the name of the SemanticTag object.
	 * 
	 * @return the name of the semantic tag
	 */
	public String getName() {
		return this.name;
	}
	
	/**
	 * Returns the meaning of the SemanticTag.
	 * 
	 * @return meaning of the semantic tag
	 */
	public String getComment() {
		return this.comment;
	}
	
	/**
	 * Returns a list of all possible parents for this SemanticTag object.
	 * 
	 * @return a list of possible parents
	 */
	public Vector getParents() {
		return this.parents;
	}
	
	/**
	 * Returns a list of all allowable children tags.
	 * 
	 * @return a list of all children tags
	 */
	public Vector getChildrenTags() {
		return this.childrenTags;
	}			
	
	/**
	 * Adds a parent to the parents list; the assumption is that 
	 * this parent is not present in the list already.
	 * 
	 * @param parent name of tha parent to be added to the list
	 */
	public void addParent(String parent) {	
		parents.addElement(parent);
	}
	
	/**
	 * Checks if this SemanticTag has "parent" in its list of parents.
	 * 
	 * @param parent the name of the tag to be check if it's a parent of this tag
	 * @return Returns true if succedes; false otherwise.
	 */
	public boolean hasParent(String parent) {
		for ( int i = 0; i < this.parents.size(); i++ ) {
			if ( ((String)parents.elementAt(i)).equals(parent) )
				return true;
		}
		return false;
	}			
	
	/**
	 * Adds <b>only the name</b> of a child to the list of children.
	 * 
	 * @param child the name of the child
	 */
	public void addChild(String child) {
		childrenTags.addElement(child);
	}
	
	/**
	 * Adds a list of children at the end of the (existing) list associated with that tag.
	 * 
	 * @param children a list of children
	 */
	public void addChildren(Vector children) {
		int i;
		for (i = 0; i < children.size(); i++)
			childrenTags.addElement((String)children.elementAt(i));		
	}	
	
	/**
	 * Adds a comment about the meaning of a tag.
	 * 
	 * @param comment comment explaining the meaning of the tag.
	 */
	public void addComment(String comment) {
		this.comment = comment;
	}
		
	/**
	 * Prints all elements in a list to the standard output.
	 * (The return type is a hack; there was a problem w/ void...)
	 * 
	 * @param list a list to be printed
	 */
	public String printList(Vector list) {
		
		if (list.size() == 0)
			System.out.println("empty list");
		else {
			int j;
			for ( j = 0; j < list.size() - 1; j++ )
				System.out.print((String)list.elementAt(j) + ", ");
			System.out.print((String)list.elementAt(j));
			System.out.println();
		}
		return "";
	}				
	
	/**
	 * Standard toString method for the SemanticTag class.
	 */
	public String toString() {		
		return ("\nname: " + this.getName() + 
				"\n\tParent(s): " + this.getParents() + 
				"\n\tChildren: " + this.getChildrenTags() + 
				"\n\tComment: " + this.getComment() + "\n");
	}	
}

